/*
 Copyright (C) 2017 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file

   \details PURPOSE: The macro produces a compressed file (date time stamped) for the user executing the macro. The compressed file consists of:
                     1. xlsx listing of job flow instances created by/shared with the logged-in user.
                     2. For modified instances (instances with uploaded data), the uploaded data sets are copied over into folders (identified with instance key) within respective librefs.

                     The feature provides a way to backup all instance specific information in a compressed directory so as to be able to restore the calculated results on a
                     different machine of the IRM platform.
                     
                     The macro must be executed in a SAS Studio session.
                        
            USAGE:   %irm_bkup_instances(debug=, logOptions=, bkup_file_path=); Must be executed in a SAS Studio session.

   <b> Identified Inputs </b>
   
   \param[in]  debug          :  True/False. If True, debugging information is printed to the log (Default: false).
   
   \param[in]  logOptions     :  Logging options such as mprint mlogic symbolgen ...
   
   \param[in]  bkup_file_path :  Path where the backup zip file will be created. The location must have write access for the logged-in user.
   
    \ingroup api
   
    \author SAS Institute INC.
    \date 2017
    
 */

%macro irm_bkup_instances(
      debug=false
   ,  logOptions=
   ,  bkup_file_path=
);
   
   %put ENTERING &sysmacroname.;
   
   /* Declare macro variables */
   %local
      username
      baseurl
      proto
      srvnm
      srvport
      server
      host
      port
      home_dir
      delim
      user
      token2
   ;
   
   %let username=&_CLIENTUSERNAME.; *Logged-in user;
   %let baseurl=&_BASEURL.; *Server URL;
   %let proto=%scan(&baseurl.,1,':');
   %let srvnm=%scan(&baseurl.,2,':/');
   
   /* Check if URL contains port */
   %let token2=%scan(&baseurl.,2,'/');

   %if %index(&token2.,:) %then %do;
      /* Extract port */
      %let srvport=%scan(&baseurl.,3,':/');
   %end;
   %else %do;
      %local SWCNAME;
      %let SWCNAME=IRM Mid-Tier Server;
      
      /* Do the search and add */
      data _null_ ;
          length rc 8 con_uri $255 port $255;
      
          /* Find our TCP/IP connection */
          con_uri="";
          rc = METADATA_GETNOBJ("omsobj:TCPIPConnection?@Service='/SASIRMServer'",1,con_uri);
          if rc ne 1 then do;
              put "ERROR: Could not locate TCP IP connection URI for '/SASIRMServer'";
              put "ERROR: Return code =" rc;
              stop;
          end;
   
          port = "";
          rc = METADATA_GETATTR(con_uri,"Port", port);
          if rc ne 0 then do;
              put "ERROR: Could not get port attribute";
              put "ERROR: Return code =" rc;
              stop;
          end;
          call symputx("srvport", trim(left(port)));
      run;
   %end;

   %let server=SASIRMServer; 
   
   %let host=&proto.%str(://)&srvnm.;
   %let port=&srvport.;
   
   %if(%length(&port.)=0) %then %do;
      %let port = 7980;
   %end;
   
   /* Store backup path in a macro variable */
   %let home_dir=%str(&bkup_file_path.);

   /* Set the required log options */
   %if(%length(&logOptions.)) %then %do;
      options &logOptions.;
   %end;

   %let delim=%str(/);
   /* Set path separator */
   %if %length(&bkup_file_path.) ne 0 %then %do;
      %if %index(&bkup_file_path,\) %then 
         %let delim=%str(\);
   %end;
   %else %do;
      %let rsk_term_message=Backup directory path not provided.;
      %put ERROR: &rsk_term_message.;
      %rsk_report_error(ERROR=No value provided for parameter bkup_file_path.);       
      %return;
   %end;
      
   /* Make sure directory exists */
   %local rc;
   %let rc=%rsk_dir_exists(dir=&bkup_file_path.);
   
   %if &rc. ne 1 %then
      %rsk_terminate(key=rsk_dir_not_found,s1=&bkup_file_path.);  
      
   /* Retrieve information about existing instances in the instances_info dataset */
   %irm_rest_get_instances(
         host=&host. 
      ,  server=&server.
      ,  port=&port. 
      ,  outds=instances_info
      ,  debug=&debug.
      ,  logOptions=&logOptions. 
   );
   
   %let user=%trim(%left(&username.));

   %local oldOptions;
   %if(%rsk_dsexist(instances_info)) %then %do;
      /* Create a backup directory and append current date time to the name. */
      %let oldOptions=%sysfunc(getoption(DLCREATEDIR)) %sysfunc(getoption(NOQUOTELENMAX));

      options dlcreatedir noquotelenmax;

      %local
         dt
         tm
         dir_nm
         work_dir
         bkup_lib
         bkup_path
      ;
      
      %let dt=%sysfunc(today(), YYMMDDN8.);
      %let tm=%sysfunc(TRANSLATE(%sysfunc(time(),NLTIME.),'-',':'));
      %let dir_nm=bkup_inst_&dt._&tm;
         
      /* Create the directory at the WORK library location */
      %let work_dir=%sysfunc(pathname(WORK));
      %let bkup_lib=libname bkup "&work_dir.&delim.&dir_nm";
      %rsk_libname(stmt=&bkup_lib.);

      %let bkup_path=%sysfunc(pathname(bkup));
      libname bkup clear;
        
      /*  Filter out instances not created by the user running the script but shared by another user.
          If retained, duplicate instance by the current user may be created.
          User running the script could be an internal user (@saspw), local user on the machine or an external domain account.
          If internal, the username contains a @saspw suffix. However, instances created by an internal user do not contain the
          @saspw suffix in created_by column in jobflow_instance. But the username with @saspw suffix is required for REST API authorization.
          Same situation arises for a domain user such as user1@<domain>. The username is stored as user1 but user1@<domain> is required for
          authentication. The possibilities are listed below:
          1. irmuser1 (a regular local user) - DB stores as irmuser1
          2. irmuser1@saspw (an internal user) - DB stores as irmuser1
          3. irmuser1@test (a regular local user) - DB stores as irmuser1@test
          4. irmuser1@test@saspw (an internal user) - DB stores as irmuser1@test
          5. irmuser1@somedomain (a domain user) - DB stores as irmuser1
          6. irmuser1@test@somedomain (a domain user) - DB stores as irmuser1@test
   
          The logic used here is to strip everything from and including the last @. This will satisfy all scenarios except for #3. 
      */ 
      
      %local 
         srch_chr
         user_int
         excel_file_nm
         bkup_title
      ;
      
      %let srch_chr=@; /* search character */
         
      %if %index(&user., &srch_chr.) gt 0 %then %do;
         /* contains @ */
         %let user_int=%substr(&user., 1, %sysfunc(findc(&user., &srch_chr., B))-1);

         /* strip out everything from and including the last @ */
         /* Retain instances created by the user. */
         data instances_info_fltrd;
            set instances_info(where=(upcase(createdBy)=upcase("&user_int.")));
         run;
      %end;
      %else %do; /* not an internal user */
         /* Retain instances created by the user. */
         data instances_info_fltrd;
            set instances_info(where=(upcase(createdBy)=upcase("&user.")));
         run;
      %end;
      
      /* Sort the dataset by instance name. */
      proc sort data=instances_info_fltrd;
         by name;
      run;
      
      /* Store excel file name in a macro variable */
      %let excel_file_nm=existing_instances_&user..xlsx;
      
      /* Retrieve title for xlsx file */
      %let bkup_title=%rsk_get_msg(key=rsk_bkup_title,s1=&user.);
      
      /* Print existing instances info to a xlsx file at the backup location. */
      ods html close;
      ods listing close;
      ods excel file="&bkup_path.&delim.&excel_file_nm." 
      options(sheet_label="Instances");
      title "&bkup_title.";
   
      proc print data=instances_info_fltrd;
         format baseDate $8.; /* Make sure base date is saved with leading zero, if any, of the format MMDDYYYY */
      run;

      /* Close the ODS destination */
      ods excel close;
      
      /* Subset list of instances based on status codes and modified input. */
      /*******************************************************
       Status Codes:
       2 = RUNNING            -Process or task is running
       3 = COMPLETED_ERROR    -Completed with error
       4 = COMPLETED_SUCCESS  -Successfully completed
       6 = PUBLISHED          -Published
       8 = PUBLISHING         -Publishing
      ********************************************************/
      /* Store instances with modified inputs in a separate data set */
      data modified_instances_info;
         set instances_info_fltrd(where=(status in (4,6) and upcase(inputDataModified)='TRUE')) ; /* already filtered for the user */
      run;

      /* Check if there are modified instances and if so, fetch the modified inputs. */
      %if %rsk_attrn(modified_instances_info, NOBS) eq 0 %then %do;
         %if (%upcase(&debug.) eq TRUE) %then %do;
            %put --------------------------------------;
            %put NO INSTANCES WITH MODIFIED INPUTS.;
            %put --------------------------------------;
         %end;            
      %end;
      %else %do;
         %local numInstances;
         %let numInstances=%rsk_attrn(modified_instances_info, NOBS);
         
         %if (%upcase(&debug.) eq TRUE) %then %do;
            %put ------------------------------------------------;
            %put THERE ARE [&numInstances.] MODIFIED INSTANCES.;
            %put ------------------------------------------------;
         %end;
         
         /* Go over the dataset and invoke the table map macro for each instance to 
            know which datasets were modified and uploaded. */
         %do k=1 %to &numInstances;
            data _null_;
               set modified_instances_info(firstobs=&k. obs=&k.);
               call symputx('instKey',key);
            run;              
      
            %irm_rest_get_table_map(
                  instanceKey=&instKey
               ,  recursive=true
               ,  host=&host.
               ,  LogonPort=
               ,  server=&server.
               ,  port=&port.
               ,  authMethod=token
               ,  outds=jobflow_table_map_&instKey.
               ,  debug=&debug.
               ,  logOptions=&logOptions.
            );
                              
            %if(%rsk_dsexist(jobflow_table_map_&instKey.)) %then %do;
               /* Check if there are rows */
               %if %rsk_attrn(jobflow_table_map_&instKey., NOBS) ne 0 %then %do;

                  /* Retain only the rows with modified inputs. */
                  data mod_inputs_&instKey.;
                     set jobflow_table_map_&instKey.(where=(upcase(modified)='TRUE')); 
                  run;

                  /* Create a folder for the Instance Key and under it individual libname folders. */
                  proc sql noprint;
                     select distinct "&bkup_path.&delim.&instKey.&delim."||trim(libname) 
                        into :lib_nm separated by ','
                        from mod_inputs_&instKey.;
                  quit;

                  %local res_lib;
                  %let res_lib=libname res ("&bkup_path.&delim.&instKey.");
                  %rsk_libname(stmt=&res_lib.);
                  libname res clear;

                  /* Fetch unique pairs of libname and dataset name into a data set and create macro variable lists. */
                  proc sql noprint;
                     create table unique_lib_data as 
                        select distinct libname, filename
                           from mod_inputs_&instKey.;
                  quit;

                  %local numRows;
                  %let numRows=&SQLOBS.;

                  proc sql noprint;
                     select libname, filename
                        into :lib1-:lib&numRows., :fl1-:fl&numRows.
                        from unique_lib_data;
                  quit;

                  /* Construct the webdav prefix to append to lib path (starts at base date). 
                  The entire WebDAv lib path is of the form: 
                  http://<server_name>:<port>/SASIRMServer/irmwebdav/<userid>/jobflows/<base_date>/<config_set_id>/<entity>/<entity_role>/<flow_category>/<flow_bpmn_name>/<instance_name>/data/<libref> */

                  /* Check for internal user and extract the name before @ if found */
                  %let srch_chr=@; /* search character */

                  %if %index(&user., &srch_chr.) gt 0 %then %do; /* internal user */ 
                     %let user=%substr(&user., 1, %sysfunc(findc(&user., &srch_chr.))-1);
                  %end;

                  %local webdavpre;
                  %let webdavpre=&host.:&port./SASIRMServer/irmwebdav/&user./jobflows/;

                  /* Loop through the macro variable list, get the webDavLibPath for a libname-filename pair, 
                     establish libref and copy the dataset over to the backup location. 
                     The webDavLibPath variable contains a deep path to the node. This may cause errors on Windows systems that have ~ 260 character limit for paths.
                     So, we substring the path until the instance name and append /data/libname to that path to access the modified input. */
                  %do i=1 %to &numRows.;
                     data _null_;
                        set mod_inputs_&instKey.(where=(libname="&&lib&i." and filename="&&fl&i.")) ;
                        /* Find the position of the 7th delimiter, which is the webdav path until instance name */
                        call scan(webDavLibPath,8,pos,len,'/\');
                        pos=pos-1;
                        path_sep=ksubstr(webDavLibPath,pos,1);
                        call symputx("dataPath&i.",ksubstr(webDavLibPath,1,pos)||'data'||trim(left(path_sep))||"&&lib&i.");
                        stop;
                     run;
                     
                     %if (%upcase(&debug.) eq TRUE) %then %do;
                        %put ------------------------------------------------;                     
                        %put WebDAV RELATIVE PATH = %bquote(&&dataPath&i.);
                        %put ------------------------------------------------;               
                     %end;

                     /* Construct complete WebDAV Library path. */
                     %local webDavPath;
                     %let webDavPath=%bquote(&webdavpre.&&dataPath&i.);

                     %if (%upcase(&debug.) eq TRUE) %then %do;
                        %put ------------------------------------------------;                     
                        %put WebDAV Complete Path = %bquote(&webDavPath.);
                        %put ------------------------------------------------;                     
                     %end;

                     /* Establish libref to source */
                     %local 
                        src_lib
                        tgt_lib
                     ;
                     %let src_lib=libname src "&webDavPath." WEBDAV AUTHDOMAIN="DefaultAuth" USER="&user.";
                     %rsk_libname(stmt=&src_lib.);

                     /* Establish libref to target */
                     %let tgt_lib=libname tgt "&bkup_path.&delim.&instKey.&delim.&&lib&i.";
                     %rsk_libname(stmt=&tgt_lib.);

                     %if %index("&&fl&i.", .) gt 0 %then %do;
                        %let fl&i.=%substr(&&fl&i., 1, %sysfunc(findc(&&fl&i., '.'))-1); 
                     %end;

                     /* Copy over the dataset */
                     proc copy out=tgt in=src;
                        SELECT &&fl&i.;
                     run;

                     /* Clear the librefs */
                     libname src clear;
                     libname tgt clear;               

                  %end; /* End of %do i=1 %to &numRows */
               %end; /* End of %if %rsk_attrn(jobflow_table_map_&instKey., NOBS) ne 0 %then %do */               
            %end; /* End of %if(%rsk_dsexist(jobflow_table_map_&instKey.)) %then %do */
            
         %end; /* End of %do k=1 %to &numInstances */
         
      %end; /* End of %if(%rsk_dsexist(modified_instances_info)) %then %do; */
        
      /* Zip-up the backup directory */
      %local zip_nm;
      %let zip_nm=&dir_nm..zip;
   
      /* Write out the directory structure in a temporary file */
      filename fil temp;
      %rsk_zip_dir(
            ods_pkg=bkup
         ,  inc_filrf=fil
         ,  dir=&bkup_path.
         ,  dir_nm=&dir_nm.
         ,  ext=sas7bdat
      );
   
      /* Create the zip file using ODS PACKAGE */
      ods package(bkup) open nopf;
      /* Add the .xlsx file */
      ods package(bkup) add file="&bkup_path.&delim.&excel_file_nm.";
      /* Add other files and directories. Check that both fileref and physical file exist */
      %if %sysfunc(fileref(fil)) eq 0 %then %do;
         %include fil;
      %end;
      ods package(bkup) 
         publish archive
            properties
               (
                  archive_name="&zip_nm."
                  archive_path="&home_dir."
               );
      ods package(bkup) close;

   %end; /* End of %if(%rsk_dsexist(instances_info)) %then %do; */
   
   /* Reset options */
   options &oldOptions.;

   %put EXITING &sysmacroname.;
%mend;
