/*
 Copyright (C) 2017 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief  This macro re-creates jobflow instances (regular and modified) based on instance information contained in a source backup directory created
           by the backup instance process.
           The combination of the backup and restore instance macros serve as a content promotion tool for IRM jobflow instances. 
           The macro must be executed in a SAS Studio session.

   <b> Identified Inputs </b> 

   \param [in] bkup_dir       Complete path (including directory name or zip file name) of the directory created by the backup instance process.
   \param [in] debug          Amount of execution information in the log. Defaulted to FALSE. Change to FALSE if DEBUG info is not required in the log.
   \param [in] logOptions     Amount of macro execution information in the log. Defaulted to blank. Could be changed to mprint mlogic symbolgen and other SAS supported options.
   \param [in] pollInterval   Number of seconds between checks for status of instance creation. Defaulted to 10 seconds.
   \param [in] maxWait        Maximum seconds to wait for an instance creation to complete. Default to 3600 seconds.        
   
   \details 
            PURPOSE: The macro creates job flow instances for the user executing the macro. For modified instances, uploaded inputs are applied to the
                     instances. Also, if instances were shared or published earlier, they will be shared or published after instance creation. 
                     The feature provides a way to re-create jobflow instances and thereby restore the calculated results on a different machine
                     of the IRM platform.
            
            PRE-REQUISITES:
                     1. Required backup directory (zipped or unzipped) is created from the backup instance process and made available to the macro.
                     2. All required federated area(s), inputs and properties for instance creation exist on the machine where instances 
                        will be restored.
 
            USAGE:   %irm_restore_instances(bkup_dir=, debug=, logOptions=, pollInterval=, maxWait=); Must be executed in a SAS Studio session.

            OUTPUT:
                     1. Creates new jobflow instances based on specifications in source directory xlsx file
                     2. Applies uploaded inputs to modified instances
                     3. PUBLISHes and/or SHAREs instances based on original instance specifications
                     4. Creates a listing of newly created instances in an xlsx file named as latest_instances_<username>.xlsx and stores the xlsx file 
                        at the location of the source backup directory in a date-time stamped folder named as restore_inst_YYYYMMDD_HH-MI-SS.
      
   \ingroup api
   
   \author  SAS Institute Inc.
   \date    2017
*/

%macro irm_restore_instances(
         bkup_dir       =
      ,  debug          = false
      ,  logOptions     =
      ,  pollInterval   = 10
      ,  maxWait        = 3600
);
                                   
   %put ENTERING &sysmacroname.; 

   /* Declare macro variables */
   %local 
      username
      baseurl
      proto
      srvnm
      srvport
      server
      host
      port
      delim
      user
      token2
   ;
   
   %let username=&_CLIENTUSERNAME.; *Logged-in user;
   %let baseurl=&_BASEURL.; *Server URL;
   %let proto=%scan(&baseurl.,1,':');
   %let srvnm=%scan(&baseurl.,2,':/'); *Host name;

   /* Check if URL contains port */
   %let token2=%scan(&baseurl.,2,'/');

   %if %index(&token2.,:) %then %do;
      /* Extract port */
      %let srvport=%scan(&baseurl.,3,':/');
   %end;
   %else %do;
      %local SWCNAME;
      %let SWCNAME=IRM Mid-Tier Server;
      
      /* Do the search and add */
      data _null_ ;
          length rc 8 con_uri $255 port $255;
      
          /* Find our TCP/IP connection */
          con_uri="";
          rc = METADATA_GETNOBJ("omsobj:TCPIPConnection?@Service='/SASIRMServer'",1,con_uri);
          if rc ne 1 then do;
              put "ERROR: Could not locate TCP IP connection URI for '/SASIRMServer'";
              put "ERROR: Return code =" rc;
              stop;
          end;
   
          port = "";
          rc = METADATA_GETATTR(con_uri,"Port", port);
          if rc ne 0 then do;
              put "ERROR: Could not get port attribute";
              put "ERROR: Return code =" rc;
              stop;
          end;
          call symputx("srvport", trim(left(port)));
      run;
   %end;
   
   %let server=SASIRMServer; 
   
   %let host=&proto.%str(://)&srvnm.; *For REST calls;
   %let port=&srvport.;
   
   %if(%length(&port.)=0) %then %do;
      %let port = 7980;
   %end;
   
   /* Set the required log options */
   %if(%length(&logOptions.)) %then %do;
      options &logOptions.;
   %end;

   %let delim=%str(/);
   /* Set path separator */
   %if %index(&bkup_dir,\) %then %do;
      %let delim=%str(\);
   %end;

   %let user=%trim(%left(&username.));
   /* Store .xlsx file name (without file extension) of the form existing_instances_<user> in a macro variable. The file contains a list of user instances. */
   %local xlsx_fil;
   %let xlsx_fil=existing_instances_&user.;

   /* Check if the backup directory path contains a .zip file */
   %local bkup_fil_nm;
   %let bkup_fil_nm=%scan(&bkup_dir.,-1,'/\');

   %local zip_file_flg;
   %let zip_file_flg=N;
   %if %index(%lowcase(&bkup_fil_nm.),.zip) gt 0 %then %do; /*Zip file found */
      %let zip_file_flg=Y;
      /* Verify that the directory location is valid */
      filename zip_fil "&bkup_dir.";
      %if %sysfunc(fileref(zip_fil)) ne 0 %then
         %rsk_terminate(key=rsk_dir_not_found,s1=&bkup_dir.);

      /* Clear the fileref */
      filename zip_fil clear;

      /* Use filename ZIP access to read the "members" (files) from the ZIP file */
      filename inzip ZIP "&bkup_dir.";

      /* Create a data set containing zip contents */
      data dir_list(keep=memname isFolder);
         length memname $1000 isFolder 8;
         fid=dopen("inzip");
         if fid=0 then
            stop;
         memcount=dnum(fid);
         do i=1 to memcount;
            memname=dread(fid,i);
            /* Check for trailing '/\' in folder name */
            isFolder = (first(reverse(trim(memname)))="&delim.");
            output;
         end;
         rc=dclose(fid);
      run;

      %local xlsx_found;
      %let xlsx_found=N;
      data _null_;
         set dir_list(where=(lowcase(memname) contains "&xlsx_fil..xlsx" and isFolder=0));
         if _n_ > 0 then
            call symputx('xlsx_found','Y');
      run;

      %if (&xlsx_found. eq Y) %then %do;
         /* Identify a temp folder in the WORK directory */
         filename xl "%sysfunc(pathname(work))&delim.&xlsx_fil..xlsx" ;
          
         data _null_;
            /* Using member syntax here */
            infile inzip(&xlsx_fil..xlsx) 
                lrecl=1024 recfm=F length=length eof=eof unbuf;
            file xl lrecl=1024 recfm=N;
            input;
            put _infile_ $varying1024. length;
            return;
          eof:
            stop;
         run;
          
         /* Convert to a SAS data set */
         proc import datafile=xl dbms=xlsx out=&xlsx_fil. replace;
         run;

         /* Clear fileref */
         filename xl;
      %end;
      %else %do; /* Terminate */
         %rsk_terminate(key=rsk_bkup_fil_not_found,s1=&xlsx_fil..xlsx,s2=&bkup_dir.);
      %end;
   %end; /* End of zip file found */
   %else %do; /* Check if an unzipped directory has been provided instead */
      /* Assigns a fileref to the directory and opens the directory */ 
      %local
         rc
         did
      ;
      %let rc=%sysfunc(filename(filrf,&bkup_dir.));                                                                                               
      %let did=%sysfunc(dopen(&filrf));                                                                                                     
                                                                                                                                        
      /* Make sure directory can be opened */                                                                                                 
      %if &did eq 0 %then                                                                                                              
         %rsk_terminate(key=rsk_dir_not_found,s1=&bkup_dir.); 

      /* Write out the directory structure in a temporary file */
      filename fil temp;
      %rsk_dir_list(
            inc_filrf=fil
         ,  dir=&bkup_dir.
      );
      
      /* Convert directory contents into a SAS data set */
      %if %sysfunc(fileref(fil)) eq 0 %then %do;
         data dir_list;
            infile fil delimiter=',';
            input memname : $2048. isFolder;
         run;
      %end;
      %else %do;
         %rsk_terminate(key=rsk_dir_error,s1=&bkup_dir.);
      %end;

      /* Look for xlsx file */
      %local xlsx_found;
      %let xlsx_found=N;
      data _null_;
         set dir_list(where=(lowcase(memname) contains "&xlsx_fil..xlsx" and isFolder=0));
         if _n_ > 0 then do;
            call symputx('xlsx_found','Y');
            call symputx('xlsx_path',memname);
         end;
      run;

      %if (&xlsx_found. eq Y) %then %do;
         /* Convert to a SAS data set */
         proc import datafile="&xlsx_path." dbms=xlsx out=&xlsx_fil. replace;
         run;
      %end;
      %else %do; /* Terminate */
         %rsk_terminate(key=rsk_bkup_fil_not_found,s1=&xlsx_fil..xlsx, s2=&bkup_dir.);
      %end;
   %end;

   /* Check if any duplicate instance exists on the target server and if found remove them from the list of instances for re-creation */
   %irm_rest_get_instances(
         host=&host. 
      ,  server=&server.
      ,  port=&port. 
      ,  outds=tgt_inst 
      ,  debug=&debug.
      ,  logOptions=&logOptions. 
   );
   
   %if(%rsk_dsexist(tgt_inst)) %then %do;
      %if %rsk_attrn(tgt_inst, NOBS) ne 0 %then %do;
         /* Create a table of only source instance names and instance keys */
         data src_inst(keep=name key);
            set &xlsx_fil. (where=(status in (4,6))); /* COMPLETED_SUCCESS=4, PUBLISHED=6 */
         run;  
         
         /* Check if there are any duplicate (same name) instances */
         proc sql noprint;
            create table dup_inst as
               select distinct 
                        src.name
                     ,  src.key
                  from   
                        src_inst src
                     ,  tgt_inst tgt
                  where
                     compress(lowcase(src.name)) = compress(lowcase(tgt.name))
                  ;
         quit;
         
         /* Output duplicate instances to the log and subset source instances list */
         %if %rsk_attrn(dup_inst, NOBS) ne 0 %then %do;
            %if (%upcase(&debug.) eq TRUE) %then %do;
               %put ================================================================;
               %put THE BELOW INSTANCE(S) ALREADY EXIST(S) AND WILL NOT BE CREATED: ;
               %put ================================================================;
               data _null_;
                  set dup_inst;
                  put 'INSTANCE NAME = ' name;
               run;
               %put ==============================================================;
            %end;
            
            proc sql noprint;
               delete from &xlsx_fil.
                  where key in (select key from dup_inst);
            quit;
         %end;
         
      %end;
   %end;

   /* Retain only successful instances (status=4). Store modified instances in another data set. */
   data all_inst mod_inst;
      set &xlsx_fil. (where=(status in (4,6))); /* COMPLETED_SUCCESS=4, PUBLISHED=6 */
      if (upcase(inputDataModified)='TRUE') then do;
         output mod_inst all_inst;
      end;
      else do;
         output all_inst;
      end;
   run;

   /* Fetch the number of flow instances that have to be created */
   %local
      numTotInst
      numModInst
   ;
   
   %let numTotInst = %rsk_attrn(all_inst, NOBS); /* Total no. of instances */
   %let numModInst = %rsk_attrn(mod_inst, NOBS); /* No. of modified instances */

   %if (%upcase(&debug.) eq TRUE) %then %do;
      %if (&numTotInst. eq 0) %then %do;
         %put --------------------------------------;
         %put NO INSTANCE FOR CREATION. EXITING;
         %put --------------------------------------;
         %return;
      %end;
      %else %do;
         %put ------------------------------------------------;
         %put NUMBER OF INSTANCES FOR CREATION = &numTotInst. ;
         %put ------------------------------------------------;
         
         %if (&numModInst. ne 0) %then %do;
            %put -----------------------------------------------;
            %put NUMBER OF MODIFIED INSTANCE(S) = [&numModInst.];
            %put -----------------------------------------------;
         %end;
      %end;
   %end; 

   %local oldOptions;
   %let oldOptions=%sysfunc(getoption(DLCREATEDIR));

   options dlcreatedir;

   /* Get the path to WORK location */
   %local 
      work_path
      NewInstanceKey
      data_key
   ;
   %let work_path=%sysfunc(pathname(work));
   %do j=1 %to &numTotInst.;
      /* Create macro variables for instance attributes. */
      data _null_;
         set all_inst(firstobs=&j. obs=&j.);
         /* Make sure baseDate has leading zero */
         attrib baseDate1 length=$8.;
         baseDate1=compress(baseDate);
         if (length(baseDate1) lt 8) then do;
            baseDate1='0'||baseDate1;
         end;
         call symputx('instanceKey',key);
         call symputx('faId',federatedAreaId);
         call symputx('configset',configSetId);
         call symputx('entityId',entityId);
         call symputx('entityRole',entityRoleKey);
         call symputx('basedate',baseDate1);
         call symputx('category',categoryId);
         call symputx('bpmnname',jobFlowDefinitionFile);
         call symputx('desc',description);
         call symputx('instanceName',name);
         call symputx('shared', visibility);
         call symputx('published', published);
         call symputx('publishComments', publishComments);
         call symputx('modInst', inputDataModified);
      run;

      /* Create the instance. Set parameters to wait for job completion. */
      %let NewInstanceKey =;
      %irm_rest_create_jobflow_instance(    
            instance_name     = %bquote(&instanceName.)
          , description       = %bquote(&desc.)
          , entityID          = &entityId.
          , entityRoleKey     = &entityRole.
          , category          = &category. 
          , jobflowFile       = &bpmnname.
          , baseDate          = &basedate.
          , configSetId       = &configset.
          , federatedAreaID   = &faId.
          , wait_flg          = Y
          , pollInterval      = &pollInterval.
          , maxWait           = &maxWait.
          , host              = &host.
          , server            = &server.
          , port              = &port.
          , outvar            = NewInstanceKey
          , logOptions        = &logOptions.
          , debug             = &debug.
      );

      %if (%upcase(&debug.) eq TRUE) %then %do;
         %put ----------------------------------------;
         %put NewInstanceKey = &NewInstanceKey.;
         %put ----------------------------------------;
      %end;            

      %if (%sysevalf(%superq(NewInstanceKey)=,boolean)) or ("%substr(&NewInstanceKey,1,2)" = "<?") %then %do;
         %rsk_report_error(error=Error in creating instance for %str(&instanceName.));
      %end;

      %if (%upcase(&modInst.) eq TRUE) %then %do;
         /* Perform processing for modified instances: Establish libref to modified input(s) and call REST macro to upload data set(s).
            Use parameters for polling and waiting before starting next execution. */

         /* Get the dataKey needed to upload input file(s) */
         %irm_rest_get_table_map(
               instanceKey = &NewInstanceKey
            ,  host = &host.
            ,  port = &port.
            ,  outds = jobflow_table_map
         );

         /* Subset dir_list data set for the instance and create macro variables for libname and table name */
         %local numMod;
         %let numMod=0;
         data _null_;
            set dir_list(where=(memname contains "&instanceKey." and isFolder=0)) end=last;
            /* Create macro variable list of the form data_path1, lib_path1, lib_nm1, tbl_nm1, data_path2, lib_path2, lib_nm2, tbl_nm2,... */
            %if (&zip_file_flg eq Y) %then %do; /* Only for a zip file source */
               call symputx('data_path'||trim(left(put(_n_,8.))),memname); /* Relative path to input data set */
            %end;
            call symputx('lib_path'||trim(left(put(_n_,8.))),substr(memname,1,findc(memname,'/\','B')-1)); /* Strip out data set name */
            call symputx('lib_nm'||trim(left(put(_n_,8.))),lowcase(scan(memname,-2,'/\')));
            call symputx('tbl_nm'||trim(left(put(_n_,8.))),lowcase(scan(memname,-2,'./\'))); /* Remove .sas7bdat from table name */

            if last then
               call symputx('numMod',_n_); /* No. of rows */
         run;

         %if &numMod. eq 0 %then %do; /* No modified input found */
            /* Write message to the log and skip uploading for the instance */
            %local mod_msg;
            %let mod_msg=%rsk_get_msg(key=rsk_input_not_found,s1=%bquote(&instanceName.));
            %put &mod_msg.;         
         %end;
         %else %do; /* Modified input found */
            /* Run a loop to convert modified inputs to xlsx and upload to the instance */
            %do i=1 %to &numMod;
               /* Fetch modified input and convert to xlsx */
               /* Check if zip file was provided */
               %if (&zip_file_flg eq Y) %then %do;
                  /* Use the inzip filref established earlier */
                  /* Create a copy of the input data set in WORK */
                  filename tl "&work_path.&delim.&&tbl_nm&i...sas7bdat";

                  data _null_;
                     /* Using member syntax here */
                     infile inzip(&&data_path&i.) lrecl=1024 recfm=F length=length eof=eof unbuf;
                     file tl lrecl=1024 recfm=N;
                     input;
                     put _infile_ $varying1024. length;
                     return;
                   eof:
                     stop;
                  run;

                  /* Create a libname to a temp folder in work location */
                  libname tgt "&work_path.&delim.datacopy";

                  /* Make a noclone copy to eliminate cross-environment data access */
                  proc copy in=work out=tgt noclone memtype=data;
                     select &&tbl_nm&i.;
                  run;

                  /* Create xlsx at WORK location */              
                  proc export data=tgt.&&tbl_nm&i.
                     outfile="&work_path.&delim.&&tbl_nm&i...xlsx"
                     dbms=XLSX REPLACE;
                  run;
               %end;
               %else %do; *Not a zip file but a folder;
                  /* Establish libref to modified input */
                  libname &&lib_nm&i. "&&lib_path&i.";

                  /* Create xlsx at WORK location */
                  proc export data=&&lib_nm&i...&&tbl_nm&i.
                        outfile="&work_path.&delim.&&tbl_nm&i...xlsx"
                        dbms=XLSX REPLACE;
                  run;
               %end;

               %let data_key=;
               /* Fetch datakey for the uploaded input from jobflow_table_map */
               data _null_;
                  set jobflow_table_map(where=(lowcase(libname)="&&lib_nm&i." and lowcase(TableName)="&&tbl_nm&i."));
                  call symputx("data_key", dataKey);
               run;

               %if &data_key. ne  %then %do;
                  /* Upload xlsx to instance */
                  %irm_rest_upload_file(
                        host           = &host.
                     ,  port           = &port.
                     ,  topInstanceKey = &NewInstanceKey.
                     ,  dataKey        = &data_key.
                     ,  dirPath        = &work_path.
                     ,  fileName       = &&tbl_nm&i...xlsx
                     ,  debug          = &debug.
                     ,  logOptions     = &logOptions.
                  );

                  /* Wait for the running instance to complete execution */
                  %irm_rest_wait_for_job(   
                        instanceKey  = &NewInstanceKey.
                     ,  host         = &host.
                     ,  port         = &port.
                     ,  pollInterval = &pollInterval.
                     ,  maxWait      = &maxWait.
                     ,  outds        = jobflow_status
                     ,  debug        = &debug.
                     ,  logOptions   = &logOptions.
                  );
               %end;
               %else %do; /* Report error */
                  %rsk_report_error(ERROR=For original instance [&instanceKey.], could not find dataKey for modified input "&&lib_nm&i...&&tbl_nm&i." for uploading to new instance [&NewInstanceKey.].);
               %end;
            %end; /* End of %do i=1 %to &numMod */
         %end;
      %end; /* End of %if (%upcase(&modInst.) eq TRUE) */

      %if (%upcase(&published.) eq TRUE) %then %do;
         /* Publish the instance. */
         %irm_rest_manage_instance(    
               instanceKey = &NewInstanceKey.
            ,  host        = &host.
            ,  server      = &server.
            ,  port        = &port.
            ,  outds       = publish_status
            ,  action      = publish
            ,  comment     = %bquote(&publishComments.)
            ,  debug       = &debug.
            ,  logOptions  = &logOptions.
         )
         
         /* Print error message if no publish privilege */
         %if(%rsk_dsexist(publish_status)) %then %do;
            %if %rsk_attrn(publish_status, NOBS) ne 0 %then %do;
               /* In case of error, the data set contains error, message and other columns */
               proc contents 
                  data=publish_status out=data_info (keep=name) noprint;
               run;

               %local pub_error;
               %let pub_error=N;
               
               data _null_;
                  set data_info;
                  if upcase(name)='ERRORCODE' then do;
                     call symputx('pub_error','Y');
                  end;
               run;

               %if (&pub_error. eq Y) %then %do;
                  data _null_;
                     set publish_status(keep=errorcode message);
                     call symputx('err_cd',errorcode);
                     call Symputx('err_msg',message);
                  run;

                  %put WARNING: [&err_cd. : %bquote(&err_msg.)];
               %end;                
            %end;
         %end;
      %end; 
      %else %do;
         %if (%upcase(&shared.) eq PUBLIC) %then %do;
            /* Share the instance. */
            %irm_rest_manage_instance(    
                  instanceKey = &NewInstanceKey.
               ,  host        = &host.
               ,  server      = &server.
               ,  port        = &port.
               ,  outds       = shared_status
               ,  action      = share
               ,  comment     =
               ,  debug       = &debug.
               ,  logOptions  = &logOptions.
            )
         %end;
      %end; /* End of %if (%upcase(&published) eq TRUE) %then %do; */ 

   %end; /* End of do loop %do j=1 %to &numTotInst; */

   /* Obtain a final list of all instances in the new_instances_info dataset. */
   %irm_rest_get_instances(
         host=&host. 
      ,  server=&server.
      ,  port=&port. 
      ,  outds=new_instances_info 
      ,  debug=&debug.
      ,  logOptions=&logOptions. 
   );
   
   %if(%rsk_dsexist(new_instances_info)) %then %do;
      /* Create a restore directory and append current date time to the name. */
      %local
         dt
         tm
         dir_nm
         src_dir_path
         restore_lib
         restore_path
         excel_file_nm
         restore_title
      ;
      
      %let dt=%sysfunc(today(), YYMMDDN8.);
      %let tm=%sysfunc(TRANSLATE(%sysfunc(time(),NLTIME.),'-',':'));
      %let dir_nm=restore_inst_&dt._&tm;
         
      /* Create the directory at the backup directory location */
      %let src_dir_path=%qsubstr(&bkup_dir.,1,%sysfunc(findc(&bkup_dir.,'/\',B))-1);
      %let restore_lib=libname restore "&src_dir_path.&delim.&dir_nm.";
      %rsk_libname(stmt=&restore_lib.);

      %let restore_path=%bquote(&src_dir_path.&delim.&dir_nm.);
      libname restore clear;

      /* Delete any pre-existing instances */
      %if(%rsk_dsexist(tgt_inst)) %then %do;
         proc sql noprint;
            delete from new_instances_info
               where name in (select name from tgt_inst);
         quit;
      %end;

      /* Sort the dataset by instance name. */
      proc sort data=new_instances_info;
         by name;
      run;
      
      /* Store excel file name in a macro variable */
      %let excel_file_nm=latest_instances_&user..xlsx;
      
      /* Retrieve title for xlsx file */
      %let restore_title=%rsk_get_msg(key=rsk_restore_title,s1=&user.);
      
      /* Print newly instances info to a xlsx file at the backup directory location */
      ods html close;
      ods listing close;
      ods excel file="&restore_path.&delim.&excel_file_nm." 
      options(sheet_label="Instances");
      title "&restore_title.";
   
      proc print data=new_instances_info;
         format baseDate $8.; /* Make sure base date is saved with leading zero, if any, of the format MMDDYYYY */
      run;

      /* Close the ODS destination */
      ods excel close;
   %end;
   
   /* Reset options */
   options &oldOptions.;

   %put EXITING &sysmacroname.;

%mend;
